# coding: utf-8

# $Id: $
import json
import logging
import socket
import sys
from datetime import datetime
import traceback


class AMQPLogstashFormatter(logging.Formatter):

    skip_list = (
        'args', 'asctime', 'created', 'exc_info', 'exc_text', 'filename',
        'funcName', 'id', 'levelname', 'levelno', 'lineno', 'module',
        'msecs', 'message', 'msg', 'name', 'pathname', 'processName',
        'relativeCreated', 'thread', 'threadName')

    def __init__(self, message_type='Logstash', tags=None, fqdn=False,
                 activity_identity={}, max_seq_no=1000):
        super(AMQPLogstashFormatter, self).__init__()
        self.message_type = message_type
        self.tags = tags if tags is not None else []
        self._seq_no = 1
        self._max_seq_no = max_seq_no
        self.ts = None

        if fqdn:
            self.host = socket.getfqdn()
        else:
            self.host = socket.gethostname()

        self.activity_identity = activity_identity
        self.skip_list = set(self.skip_list) - set(activity_identity.keys())

    def get_debug_fields(self, record):
        fields = {
            'exc_info': self.format_exception(record.exc_info),
            'lineno': record.lineno,
            'process': record.process,
            'threadName': record.threadName,
        }

        # funcName was added in 2.5
        if not getattr(record, 'funcName', None):
            fields['funcName'] = record.funcName

        # processName was added in 2.6
        if not getattr(record, 'processName', None):
            fields['processName'] = record.processName

        return fields

    @classmethod
    def format_source(cls, message_type, host, path):
        return "%s://%s/%s" % (message_type, host, path)

    @classmethod
    def format_timestamp(cls, time):
        tstamp = datetime.fromtimestamp(time)
        dt = tstamp.strftime("%Y-%m-%dT%H:%M:%S")
        msec = tstamp.microsecond / 1000
        return '%s.%03dZ' % (dt, msec)

    @classmethod
    def format_exception(cls, exc_info):
        if not exc_info:
            return ''
        exception = traceback.format_exception(*exc_info)
        return ''.join(exception)

    @classmethod
    def serialize(cls, message):
        if sys.version_info < (3, 0):
            return json.dumps(message)
        else:
            # noinspection PyArgumentList
            return bytes(json.dumps(message), 'utf-8')

    def get_seq_no(self, created):
        if created != self.ts:
            self.ts = created
            self._seq_no = 0
        if self._seq_no < self._max_seq_no:
            self._seq_no += 1
        return self._seq_no

    def format(self, record):
        # Create message dict
        message = {
            '@timestamp': self.format_timestamp(record.created),
            '@version': 1,
            'message': record.getMessage(),
            'host': self.host,
            'tags': self.tags,
            'type': self.message_type,
            'seq': self.get_seq_no(record.created),
            # Extra Fields
            'levelname': record.levelname,
            'logger': record.name,
        }

        # Add extra fields
        message.update(self.get_extra_fields(record))
        message.update(self.activity_identity)

        # If exception, add debug info
        if record.exc_info:
            message.update(self.get_debug_fields(record))

        return self.serialize(message)

    def get_extra_fields(self, record):
        # The list contains all the attributes listed in
        # http://docs.python.org/library/logging.html#logrecord-attributes

        if sys.version_info < (3, 0):
            easy_types = (basestring, bool, dict, float, int, long, list,
                          type(None))
        else:
            easy_types = (str, bool, dict, float, int, list, type(None))

        fields = {}

        for key, value in record.__dict__.items():
            if key not in self.skip_list:
                if isinstance(value, easy_types):
                    fields[key] = value
                else:
                    fields[key] = repr(value)

        return fields


# coding: utf-8

# $Id: $
from logging.handlers import SocketHandler
from urlparse import urlparse
import errno

from amqp import Message
from amqp.connection import Connection, AMQP_LOGGER


class AMQPHandler(SocketHandler, object):

    def __init__(self, broker_uri='amqp://localhost/', exchange='logstash',
                 exchange_type='topic',
                 message_type='logstash', tags=None,
                 durable=True, version=0, extra_fields=True, fqdn=False,
                 facility=None, routing_key='logstash'):
        super(AMQPHandler, self).__init__(None, None)
        self.broker_uri = broker_uri
        self.exchange = exchange
        self.routing_key = routing_key
        self.exchange_type = exchange_type
        self.durable = durable

    def emit(self, record):
        return super(AMQPHandler, self).emit(record)

    def makeSocket(self, **kwargs):
        propagate = AMQP_LOGGER.propagate
        try:
            AMQP_LOGGER.propagate = False
            amqp_socket = AMQPSocket(self.broker_uri,
                                     self.exchange,
                                     self.exchange_type,
                                     self.routing_key,
                                     self.durable)
        finally:
            AMQP_LOGGER.propagate = propagate
        return amqp_socket

    def makePickle(self, record):
        return self.formatter.format(record)


class AMQPSocket(object):

    def __init__(self, broker_uri, exchange, exchange_type, routing_key,
                 durable):
        self.is_logging = True
        self.conn = Connection(**self._parse_broker_uri(broker_uri))
        self.channel = self.conn.channel()
        self.channel.exchange_declare(exchange=exchange,
                                      type=exchange_type,
                                      durable=durable,
                                      auto_delete=not durable)
        self.is_logging = False
        self.exchange = exchange
        self.routing_key = routing_key

    def sendall(self, data):
        if self.is_logging:
            return
        self.is_logging = True
        try:
            msg = Message(data)
            self.channel.basic_publish(msg, self.exchange, self.routing_key)
        finally:
            self.is_logging = False

    def close(self):
        try:
            self.conn.close()
        except IOError as e:
            # handle already closed connections
            if e.errno != errno.EPIPE:
                raise

    @staticmethod
    def _parse_broker_uri(broker_uri):
        url = urlparse(broker_uri)
        return {
            'host': url.hostname,
            'virtual_host': url.path,
            'userid': url.username,
            'password': url.password
        }


# coding: utf-8

# $Id: $


from logging import root, _handlers, config

from celery import signals

from logcollect.formatter import AMQPLogstashFormatter
from logcollect.handler import AMQPHandler

LOGGING = {
    'version': 1,
    'disable_existing_loggers': False,
    'formatters': {
        'verbose': {
            'format': '%(levelname)s %(asctime)s %(module)s %(message)s'
        },
    },
    'handlers': {
        'console': {
            'level': 'DEBUG',
            'class': 'logging.StreamHandler',
            'formatter': 'verbose'
        }
    },
    'loggers': {
        '': {
            'level': 'DEBUG',
            'handlers': []
        }
    }
}


def default_config(broker_uri='amqp://localhost/', exchange='logstash',
                   routing_key='logstash', durable=False, level='DEBUG',
                   activity_identity={}):
    config.dictConfig(LOGGING)
    return ensure_amqp_handler(broker_uri, exchange, routing_key, durable,
                               level, activity_identity)


def ensure_amqp_handler(broker_uri, exchange, routing_key, durable, level,
                        activity_identity, logger=None):
    amqp_handler = None
    logger = logger or root
    for h in _handlers.values():
        if isinstance(h, AMQPHandler):
            amqp_handler = h
    if amqp_handler is None:
        amqp_handler = AMQPHandler(broker_uri=broker_uri,
                                   exchange=exchange, durable=durable,
                                   routing_key=routing_key)
        amqp_handler.setFormatter(AMQPLogstashFormatter(
            activity_identity=activity_identity))
        amqp_handler.setLevel(level)
    root_handler = None
    for h in logger.handlers:
        if isinstance(h, AMQPHandler):
            root_handler = h
    if not root_handler:
        logger.addHandler(amqp_handler)
    return amqp_handler


def django_dict_config(LOGGING, broker_uri='amqp://localhost/',
                       exchange='logstash', routing_key='logstash',
                       durable=False, level='DEBUG', activity_identity={}):
    LOGGING.setdefault('handlers', {})
    amqp_handler = None
    for name, handler_conf in LOGGING['handlers'].items():
        if handler_conf['class'] == 'logcollect.handler.AMQPHandler':
            amqp_handler = name

    if not amqp_handler:
        amqp_handler = 'logcollect'
        if amqp_handler in LOGGING['handlers']:
            raise ValueError("logcollect handler name conflict")
        handler_conf = {
            'class': 'logcollect.handler.AMQPHandler',
        }
        LOGGING['handlers'][amqp_handler] = handler_conf

    LOGGING.setdefault('formatters', {})
    amqp_formatter = None
    for name, formatter_conf in LOGGING['formatters'].items():
        if formatter_conf.get(
                '()') == 'logcollect.formatter.AMQPLogstashFormatter':
            amqp_formatter = name

    if not amqp_formatter:
        amqp_formatter = 'logcollect'
        if amqp_formatter in LOGGING['formatters']:
            raise ValueError("logcollect formatter name conflict")
        formatter_conf = {
            '()': 'logcollect.formatter.AMQPLogstashFormatter',
        }
        formatter_conf.setdefault('activity_identity', activity_identity)
        LOGGING['formatters'][amqp_formatter] = formatter_conf

    LOGGING.setdefault('loggers', {})
    LOGGING['loggers'].setdefault('', {})
    root_logger = LOGGING['loggers']['']
    root_logger.setdefault('handlers', [])
    if amqp_handler not in root_logger['handlers']:
        root_logger['handlers'].append(amqp_handler)
    root_logger.setdefault('level', level)

    handler_conf = LOGGING['handlers'][amqp_handler]
    handler_conf.setdefault('broker_uri', broker_uri)
    handler_conf.setdefault('exchange', exchange)
    handler_conf.setdefault('routing_key', routing_key)
    handler_conf.setdefault('durable', durable)
    handler_conf.setdefault('level', level)
    handler_conf.setdefault('formatter', amqp_formatter)

    return amqp_handler


def celery_config(broker_uri='amqp://localhost/',
                  exchange='logstash', routing_key='logstash',
                  durable=True, level='DEBUG', activity_identity={},
                  collect_root_logs=False):

    def init_logging(**kwargs):
        from celery.utils.log import task_logger
        if collect_root_logs:
            ensure_amqp_handler(broker_uri, exchange, routing_key, durable,
                                level, activity_identity, logger=None)
        ensure_amqp_handler(broker_uri, exchange, routing_key, durable,
                            level, activity_identity, logger=task_logger)

    signals.worker_process_init.connect(init_logging, weak=False)
    signals.worker_ready.connect(init_logging, weak=False)



from distutils.core import setup

try:
    from pypandoc import convert
    read_md = lambda f: convert(f, 'rst')
except ImportError:
    print("warning: pypandoc module not found, could not convert Markdown to RST")
    read_md = lambda f: open(f, 'r').read()

setup(
    name='logcollect',
    version='0.10.0',
    long_description=read_md('README.md'),
    packages=['logcollect'],
    url='http://github.com/rutube/logcollect/',
    license='Beerware',
    author='tumbler',
    author_email='zimbler@gmail.com',
    description='Helper for collecting logs to ELK stack via RabbitMQ',
    setup_requires=['amqp'],
)
